#!/usr/bin/env python
#
# Syntax Tree -> Image conversion program.
#
# This program was written for use with the x-render xemacs library.
# This allows me to view my syntax trees as I write them in my notes. :)
#
# Author: Edward Loper <edloper@gradient.cis.upenn.edu>
# URL: <http://www.cis.upenn.edu/~edloper/tree2image/>
#
# $Id: tree2image,v 1.8 2002/05/05 15:03:39 edloper Exp $

"""
Render a (syntactic) tree as an image.  C{tree2image} can be used from
the command line, or from other Python programs.

The C{tree2image} module defines three basic interfaces:

  - C{parse_tree} converts the string representation of a syntactic
    tree to a hierarchical encoding for a picture of the tree.  This
    encoding is called a "diagram."
  - C{DiagramRenderer} converts a diagram to a flat representation
    of the lines and text strings needed to draw the tree.  This
    representation is called a "shape list."
  - C{shapelist2image} and C{shapelist2ps} convert shape lists to
    raster images and postscript images, respectively.

Each of these interfaces can be used independantly of the other.
C{tree2image} also defines a X{command-line interface} for rendering a
file containing a string representation of a syntactic tree as an
image file.  Internally, the command-line interface uses C{parse_tree}
to convert the string to a diagram; C{DiagramRenderer} to convert that
diagram to a shape list; and C{shapelist2image} or C{shapelist2ps} to
convert the shape list to an image::

       string representation of tree
                    |
              [parse_tree]
                    |
                    V
                 diagram
                    |
            [DiagramRenderer]
                    |
                    V
                shape list
                /        \\  
    [shapelist2image] [shapelist2ps]
              /            \\  
             V              V 
      raster image    postscript image

The C{tree2image} module uses two intermediate representations for
pictures of tree structures:

  - A X{diagram} is a hierarchical encoding for a picture of
    a tree.  Diagrams only encode the hierarchical relationships
    between elements of the picture; they do not represent any
    absolute positional information.  Each diagram consists of a
    C{DiagramWidget}, which may contain zero or more child widgets.
    An example of a X{diagram} is:

        >>> TreeWidget('VP', 'saw', TreeWidget('NP', 'the', 'dog'))
        ['VP': 'saw' ['NP': 'the' 'dog']]

  - A X{shape list} encodes the absolute position of elementary
    graphical elements (such as lines and text strings) that make up a
    picture of a tree.  This is used as a very simple common interface
    for different image-drawing packages (e.g., PIL and Tkinter).
    Each shape list consists of a C{list} of C{Shape}s.  An example of
    a shape list is:

        >>> [LineShape(20,20,10,30), LineShape(20,20,30,30),
        ...  TextShape(20,10,'NP')]
        [(20,20-10,30), (20,20-30,30), ('NP'@20,10)]

These representations are intended to be stable in their overall
structure, but may grow as I addd more functionality to C{tree2image}.

@author: Edward Loper
"""

"""
To do:
  - better documentation
  - parser for LaTeX parsetree package?  Other packages?

Impl note:
  - we draw in RGB and then convert to mono/grayscale, rather than
    initially drawing in mono or grayscale, because solaris seems to
    ignore font colors in mono/grayscale.
"""

import sys, re

# This should eventually be made a command-line option, since some
# fonts won't support it.  If it's on, then accented characters will
# be rendered with extended ASCII.  Leave it on, since helvR08.pil
# supports it.
REPLACE_LATEX_CHARS = 1

# Default values...
DEFAULT_XSPACE = 1.5
DEFAULT_YSPACE = 1.5
DEFAULT_LINESPACE = 1.0
DEFAULT_FONT = 'helvR08.pil'
DEFAULT_PSFONT = ('times', 12)
DEFAULT_IMAGECOLOR = 'color' # can be monochrome or grayscale or color
DEFAULT_VERBOSITY = 0
DEFAULT_REVERSE_VIDEO = 0
DEFAULT_LINECOLOR = '#80ffff'
DEFAULT_TEXTCOLOR = '#ffffff'
DEFAULT_BGCOLOR = '#000000'
DEFAULT_PS_IMAGECOLOR = 'monochrome'
DEFAULT_PS_LINECOLOR = '#000000'
DEFAULT_PS_TEXTCOLOR = '#000000'
DEFAULT_PS_BGCOLOR = '#ffffff'

# This is very helpful for debugging the drawing routines:
DEBUG_DRAW = 0

# Use special postscript handling?  Experimental.
SPECIAL_PS = 1

__all__ = (
    # Diagrams:
    'DiagramWidget', 'TextWidget', 'TreeWidget',
    'SequenceWidget', 'StackWidget',
    # Shapes:
    'Shape', 'TextShape', 'LineShape',
    # String -> Diagram:
    'parse_tree', 'parse_treebank_tree', 'parse_latex_qtree',
    # Diagram -> Shapes:
    'DiagramRenderer',
    # Shapes -> Image:
    'shapelist2image', 'shapelist2ps',
    # Misc.:
    'PseudoFont', 'replace_latex_chars', 'cli',
    )
    

############################################################
##  Diagram representation
############################################################

class DiagramWidget:
    """

    A hierarchical representation for tree diagrams, which is used as
    an interface between C{parse_tree} and C{DiagramRenderer}.  Each
    C{DiagramWidget} is a single object in the diagram representing a
    tree structure.  

    C{DiagramWidget} is an abstract superclass.  Each of its
    subclasses defines a specific type of C{DiagramWidget}.

    For convenience, C{string}s can be used anywhere where a
    C{DiagramWidget} is expected; they will automatically be converted
    into C{TextWidget}s.  Thus, for example, you can use:

        >>> TreeWidget('NP', 'the', dog')
        [NP: the, dog]

    instead of:

        >>> TreeWidget(TextWidget('NP'), TextWidget('the'), TextWidget('dog'))
        [NP: the dog]

    Also, C{None} can be used wherever a C{DiagramWidget} is expected;
    it will be automatically converted to a C{TextWidget} containing
    the empty string.

    History
    =======
    The diagram representation is derived from the C{CanvasWidget}
    representation defined by NLTK (C{http://nltk.sf.net}), which is
    in turn partially based on C{CLIG}
    (C{http://www.ags.uni-sb.de/~konrad/clig.html}).  Unlike
    C{CanvasWidget} and C{CLIG} objects, C{DiagramWidget}s are only
    used to store the structures; they do not draw them.  Instead,
    C{DiagramRenderer} is responsible for drawing them.  There are
    three reasons to use C{DiagramWidget}s instead of
    C{CanvasWidget}s:
    
        1. C{CanvasWidget}s are dependant on C{Tkinter} canvases;
           since C{DiagramWidget}s contain only data (and no drawing
           instructions), they are more output-format neutral.  As a
           result, it's easier to use them with PIL or other output
           formats. 
        2. Using C{CanvasWidget}s would create a dependancy on NLTK,
           which seems unnecessary.  I wanted to keep C{tree2image}
           small and self-contained.
        3. C{DiagramWidget}s are simpler, since they implement much
           less functionality than NLTK C{CanvasWidget}s.  In
           particular, C{CanvasWidget}s implement a large amount of
           user-interaction functionality, which is not needed by
           C{tree2image}. 

    I may add new C{DiagramWidget}s in the future, as I increase the
    functionality of C{tree2image}.  Note that if any new
    C{DiagramWidget}s are added, then C{DiagramRenderer} and any other
    diagram rendering functions/classes must be changed to handle the
    new widgets.
    """
    def __init__():
        raise AssertionError('DiagramWidget is an abstract class')

    def _aswidget(self, val):
        """
        If C{val} is a C{DiagramWidget}, then return C{val};
        otherwise, try to coerce C{val} into a C{DiagramWidget}, and
        return it.  Currently, C{string}s and C{None} can be coerced
        into C{DiagramWidget}s.

        @return: C{val}, coerced into a C{DiagramWidget}
        @rtype: C{DiagramWidget}
        @raise TypeError: If C{val} cannot be coerced into a
            C{DiagramWidget}. 
        """
        if isinstance(val, DiagramWidget): return val
        elif type(val) == type(''): return TextWidget(val)
        elif val is None: return TextWidget('')
        else: raise TypeError('Expected DiagramWidget')

class TreeWidget(DiagramWidget):
    """
    A diagram widget that connects a node widget to zero or more child
    widgets.  The node widget is drawn on top, centered above the
    child widgets.  If the roof attribute is false, then a line is
    drawn to each child widget; otherwise, a triangular roof is drawn
    over all of the child widgets.
    
    Currently, the following attributes are defined:
      - C{roof} specifies whether this level of the tree should be
        drawn with a triangular roof (instead of separate lines to
        each child).

    @ivar node: The C{TreeWidget}'s node value.
    @type node: C{DiagramWidget}
    @ivar children: The list of the C{TreeWidget}'s children.
    @type children: C{list} of {DiagramWidget}
    @ivar attribs: The C{TreeWidget}'s attributes
    @type attribs: C{dictionary}
    """
    def __init__(self, node=None, *children, **attribs):
        self.node = self._aswidget(node)
        self.children = [self._aswidget(c) for c in children]
        self.attribs = attribs

    def with_node(self, node):
        """
        @return: a new C{TreeWidget} whose children and attributes are
            identical to this C{TreeWidget}, but whose node is C{node}.
            This is used to simplify parsing functions.
        @rtype: C{TreeWidget}
        """
        return TreeWidget(node, *self.children, **self.attribs)
    
    def __repr__(self):
        children = [repr(c) for c in self.children]
        return '[%r: %s]' % (self.node, ' '.join(children))

class StackWidget(DiagramWidget):
    """
    A diagram widget that places two or more sub-widgets in a vertical
    line.  Currently, the following attributes are defined:
      - C{align}: The horizontal alignment of the sub-widgets.  Possible
        values are C{'left'}, C{'center'}, and C{'right'}.  By
        default, sub-widgets are center-aligned.
      - C{space}: The amount of vertical space to place between
        sub-widgets.  By default, one pixel of space is used.

    @ivar subwidgets: The stack's subwidgets.
    @type subwidgets: C{list} of C{DiagramWidget}
    @ivar attribs: The C{StackWidget}'s attributes
    @type attribs: C{dictionary}
    """
    def __init__(self, *subwidgets, **attribs):
        self.subwidgets = [self._aswidget(sw) for sw in subwidgets]
        self.attribs = attribs

    def __repr__(self):
        return '<stack: %s>' % self.subwidgets
    
class SequenceWidget(DiagramWidget):
    """
    A diagram widget that places two or more sub-widgets in a
    horizontal line.  Currently, the following attributes are defined:
    
      - C{align}: The vertical alignment of the sub-widgets.  Possible
        values are C{'top'}, C{'center'}, and C{'bottom'}.  By
        default, sub-widgets are center-aligned.
      - C{space}: The amount of vertical space to place between
        sub-widgets.  By default, one pixel of space is used.

    @ivar subwidgets: The stack's subwidgets.
    @type subwidgets: C{list} of C{DiagramWidget}
    @ivar attribs: The C{SequenceWidget}'s attributes
    @type attribs: C{dictionary}
    """
    def __init__(self, *subwidgets, **attribs):
        self.subwidgets = [self._aswidget(sw) for sw in subwidgets]
        self.attribs = attribs

    def __repr__(self):
        return '<sequence: %s>' % self.subwidgets
    
class TextWidget(DiagramWidget):
    """
    A diagram widget containing uniformly a formatted text string.
    Currently, the following attributes are defined:
      - C{subscript}: region is subscripted
      - C{superscript}: region is superscripted

    @ivar str: The widget's text string.
    @type str: C{string}
    @ivar attribs: The C{SequenceWidget}'s attributes
    @type attribs: C{dictionary}
    """
    def __init__(self, str, **attribs):
        if type(str) != type(''): raise TypeError('TextWidget: expected string')
        self.str = str
        self.attribs = attribs

    def __repr__(self):
        return repr(self.str)

############################################################
##  Shape list representation
############################################################

class Shape:
    """
    A simple representation for a graphical element, such as a line or
    a text string, which is used as an interface between
    C{DiagramRenderer} and C{shapelist2image}.  C{Shape}s are
    intentionally very simple, to make it easy to use different
    image-drawing packages (e.g., PIL and Tkinter).

    Each C{Shape} defines a number of public instance variables, and a
    C{shift} function, which moves the object by a given distance.
    """
    def __init__(self):
        raise AssertionError('Shape is an abstract class')

    def shift(self, dx, dy):
        """
        Move this shape by a given distance.  In particular, shift the
        shape right by C{dx} pixels, and down by C{dy} pixels.  Both
        C{dx} and C{dy} may be negative, resulting in leftward or
        upward movement.

        @type dx: C{int}
        @param dx: The number of pixels to move this shape rightwards.
        @type dy: C{int}
        @param dy: The number of pixels to move this shape downwards.
        @rtype: C{None}
        """
        raise AssertionError('Shape is an abstract class')

class LineShape:
    """
    The C{Shape} encoding for lines.

    @type x1: C{int}
    @ivar x1: The x coordinate for the start of the line.
    @type y1: C{int}
    @ivar y1: The y coordinate for the start of the line.
    @type x2: C{int}
    @ivar x2: The x coordinate for the end of the line.
    @type y2: C{int}
    @ivar y2: The y coordinate for the end of the line.
    """
    def __init__(self, x1, y1, x2, y2):
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2

    def shift(self, dx, dy):
        self.x1 += dx
        self.x2 += dx
        self.y1 += dy
        self.y2 += dy

    def __repr__(self):
        return '(%d,%d-%d,%d)' % (self.x1, self.y1, self.x2, self.y2)

class TextShape:
    """
    The C{Shape} encoding for text strings.

    @type x: C{int}
    @ivar x: The x coordinate of the text string's upper-left corner.
    @type y: C{int}
    @ivar y: The y coordinate of the text string's upper-left corner.
    @type text: C{string}
    @ivar text: The text string to display.
    """
    def __init__(self, x, y, text):
        self.x = x
        self.y = y
        self.text = text

    def shift(self, dx, dy):
        self.x += dx
        self.y += dy

    def __repr__(self):
        return '(%r@%d,%d)' % (self.text, self.x, self.y)

############################################################
##  Tree String Parsing
############################################################

def parse_tree(str):
    """
    Given a string containing a representation of a syntax tree,
    determine which representation is used, and return a C{TreeWidget}
    representing the syntax tree.

    @rtype: C{TreeWidget}
    """
    str = str.strip()

    # If every line (except possibly the 1st) begins with "#", then
    # strip them off.  This is useful for my xemacs notes mode.
    lines = str.split('\n')
    if not [1 for line in lines[1:] if line[0:1] != '#']:
        str = '\n'.join([lines[0]]+[line[1:] for line in lines[1:]])
    
    if str[0:1] == '[[':
        return parse_treebank_tree(str[1:-1], '[]')
    elif str[0:1] == '((':
        return parse_treebank_tree(str[1:-1], '()')
    if str[0:1] == '[':
        return parse_treebank_tree(str, '[]')
    elif str[0:1] == '(':
        return parse_treebank_tree(str, '()')
    elif str[0:5] == '\Tree':
        return parse_latex_qtree(str)
    else:
        raise ValueError('Unknown tree format')

def parse_treebank_tree(str, delimiters='[]', latex_accented_chars=0):
    """
    Given a string containing a Treebank-style representation of a
    syntax tree, return a C{TreeWidget} representing that syntax tree.
    For example, the string::

       [IP [DP the dog] [VP [V [V' saw] [DP [D the] [NP man]]]]]

    Will be converted into the C{TreeWidget} object::

       TreeWidget('IP', TreeWidget('DP', 'the', 'dog',),
                  TreeWidget('VP', TreeWidget('V', TreeWidget('V\'', 'saw',),
                             TreeWidget('DP', TreeWidget('D', 'the',),
                                        TreeWidget('NP', 'man')))))

    Quotes and backslash are treated like normal characters.  In
    particular, you cannot quote or backslash the delimiter
    characters.  For example, the string::

      [Test \[ '[' "]" \] ]

    Will be parsed as::

      TreeWidget('Test', '\\', TreeWidget("'", TreeWidget("'", '"'),'"', '\\'))

    @param delimiters: A length-2 string specifying the open and close
      delimiter characters.  Typical values are '()' and '[]'.
    @type delimiters: C{string}
    @rtype: C{TreeWidget}
    """
    # Rather than trying to parse the string ourselves, we will rely
    # on the Python parser.  This procedure isn't immediately easy to
    # understand, but if you add a "print str" after each call to
    # re.sub, and try it out on an example, you should be able to
    # figure out what's going on.
    (L, R) = delimiters
    if L in '()[]{}': L = '\\'+L
    if R in '()[]{}': R = '\\'+R

    # Prevent delimiters from getting confused with other chars.
    str = re.sub(L, '\0', str); L = '\0'
    str = re.sub(R, '\1', str); R = '\1'

    # Handle latex accented characters?
    if latex_accented_chars:
        str = replace_latex_chars(str)
        
    # Get rid of any newlines, etc.
    str = ' '.join(str.split())
    # Backslash any quote marks or backslashes in the string.
    str = re.sub(r'([\\"\'])', r'\\\1', str)
    # Replace words with calls to parse_treebank_label
    str = re.sub('([^'+L+R+' ]+)', r"_parse_treebank_label('\1'),", str)
    # Replace left delimiter with constructor call
    str = re.sub(L, 'TreeWidget(', str)
    # Replace right delimiter with close paren
    str = re.sub(R, '),', str)
    # Strip whitespace and get rid of the comma after the last ')'
    str = str.strip()[:-1]
    # Use "eval" to convert the string (is this safe?)
    try:
        result = eval(str)
        return result
    except:
        raise
        raise ValueError('Bad Treebank-style string')

def _parse_treebank_label(str):
    lines = []
    line_strings = str.split('\\\\')
    for line_str in line_strings:
        substrings = re.split(r'([_^]\{[^{}]+\}|\^[^{]|_[^{])', line_str)
        line = []
        for substring in substrings:
            if len(substring) == 0: continue
            if substring[:2] == '^{' and substring[-1:] == '}':
                line.append(TextWidget(substring[2:-1], superscript=1))
            elif substring[:2] == '_{' and substring[-1:] == '}':
                line.append(TextWidget(substring[2:-1], subscript=1))
            elif len(substring) == 2 and substring[0] == '^':
                line.append(TextWidget(substring[1], superscript=1))
            elif len(substring) == 2 and substring[0] == '_':
                line.append(TextWidget(substring[1], subscript=1))
            else:
                line.append(TextWidget(substring))
        lines.append(SequenceWidget(*line))
            
    return StackWidget(*lines)

def _extract_latex_literals(str):
    """
    A helper function for dealing with latex literals.  Given a string
    that might contain literals of the form {...} and $...$, return a
    a new string with each literal replaced by "{n}", where n is an
    integer; and a list of the literals, where literal "{n}"
    corresponds to element n of this list.  For example::

      >>> str = 'This {is {a test} $of$} the LaTeX $literal {extractor}$'
      >>> (str, literals) = _extract_latex_literals(str)[0]
      >>> print str
      This {0} the LaTeX $literal {1}$
      >>> print literals
      ['{is {a test} $of$}', '{extractor}']
    
    """
    braces = 0 # How deep are we in 
    equation = 0 # Are we in a $...$ equation region?
    backslashed = 0 # are we backslashed?
    start = None # Start of the current region.
    i = 0 # Current index.
    literals = [] # Literals we found.  Just keep track of (start,end)

    # First, find the literals.
    for i in range(len(str)):
        c = str[i]
        
        if c == '\\': backslashed = not backslashed
        else: backslashed = 0

        if backslashed: pass
        elif c == '{': # Start braces
            if braces == 0 and not equation: start = i
            braces += 1
        elif c == '$' and not equation: # Start equation
            if braces == 0: start = i
            equation = 1
        elif c == '}': # End braces
            braces -= 1
            if braces == 0 and not equation: literals.append((start, i+1))
        elif c == '$' and equation: # End equation
            if braces == 0: literals.append((start, i+1))
            equation = 0

    # Construct a new string, with the literals removed.
    literals.append((len(str),len(str)))
    newstr = str[:literals[0][0]]
    for i in range(len(literals)-1):
        newstr += '{%d}' % i + str[literals[i][1]:literals[i+1][0]]

    return (newstr, [str[s:e] for (s,e) in literals])

def replace_latex_chars(str):
    """
    Convert accented characters from LaTeX command sequences (e.g.,
    C{\\`A} or C{\\emptyset}) with extended ASCII characters.  Note:
    this will only produce correct results if the font file used by
    tree2image has the correct entries for extended ASCII characters.
    """
    str = re.sub(r'\\`A', chr(192), str)                # 
    str = re.sub(r"\\'A", chr(193), str)                # 
    str = re.sub(r'\\\^A', chr(194), str)               # 
    str = re.sub(r'\\~A', chr(195), str)                # 
    str = re.sub(r'\\"A', chr(196), str)                # 
    str = re.sub(r'\\\^c C', chr(199), str)             # 
    str = re.sub(r'\\`E', chr(200), str)                # 
    str = re.sub(r"\\'E", chr(201), str)                # 
    str = re.sub(r'\\\^E', chr(202), str)               # 
    str = re.sub(r'\\"E', chr(203), str)                # 
    str = re.sub(r'\\`I', chr(204), str)                # 
    str = re.sub(r"\\'I", chr(205), str)                # 
    str = re.sub(r'\\\^I', chr(206), str)               # 
    str = re.sub(r'\\"I', chr(207), str)                # 
    str = re.sub(r'\\~N', chr(209), str)                # 
    str = re.sub(r'\\`O', chr(210), str)                # 
    str = re.sub(r"\\'O", chr(211), str)                # 
    str = re.sub(r'\\\^O', chr(212), str)               # 
    str = re.sub(r'\\~O', chr(213), str)                # 
    str = re.sub(r'\\"O', chr(214), str)                # 
    str = re.sub(r'\\times', chr(215), str)             # 
    str = re.sub(r'\\O|\\emptyset', chr(216), str)      # 
    str = re.sub(r'\\`U', chr(217), str)                # 
    str = re.sub(r"\\'U", chr(218), str)                # 
    str = re.sub(r'\\\^U', chr(219), str)               # 
    str = re.sub(r'\\"U', chr(220), str)                # 
    str = re.sub(r"\\'Y", chr(221), str)                # 
    str = re.sub(r'\\`a', chr(224), str)                # 
    str = re.sub(r"\\'a", chr(225), str)                # 
    str = re.sub(r'\\\^a', chr(226), str)               # 
    str = re.sub(r'\\~a', chr(227), str)                # 
    str = re.sub(r'\\"a', chr(228), str)                # 
    str = re.sub(r'\\c c', chr(231), str)               # 
    str = re.sub(r'\\`e', chr(232), str)                # 
    str = re.sub(r"\\'e", chr(233), str)                # 
    str = re.sub(r'\\\^e', chr(234), str)               # 
    str = re.sub(r'\\"e', chr(235), str)                # 
    str = re.sub(r'\\`i', chr(236), str)                # 
    str = re.sub(r"\\'i", chr(237), str)                # 
    str = re.sub(r'\\\^i', chr(238), str)               # 
    str = re.sub(r'\\"i', chr(239), str)                # 
    str = re.sub(r'\\~n', chr(241), str)                # 
    str = re.sub(r'\\`o', chr(242), str)                # 
    str = re.sub(r"\\'o", chr(243), str)                # 
    str = re.sub(r'\\\^o', chr(244), str)               # 
    str = re.sub(r'\\~o', chr(245), str)                # 
    str = re.sub(r'\\"o', chr(246), str)                # 
    str = re.sub(r'\\o', chr(248), str)                 # 
    str = re.sub(r'\\`u', chr(249), str)                # 
    str = re.sub(r"\\'u", chr(250), str)                # 
    str = re.sub(r'\\\^u', chr(251), str)               # 
    str = re.sub(r'\\"u', chr(252), str)                # 
    str = re.sub(r"\\'y", chr(253), str)                # 
    str = re.sub(r'\\"y', chr(255), str)                # 

    # Get rid of extraneous braces
    str = re.sub('{([\x80-\xff])}', r'\1', str)
    
    return str
    
def parse_latex_qtree(str):
    """
    Given a string containing a LaTeX qtree-mode representation of a
    syntax tree, return a C{TreeWidget} representing that syntax tree.
    For example, the string::

       \Tree [.IP \qroof{the dog}.DP
                  [.VP [.V [.V' saw ] [.DP [.D the ] [.NP man ]]]]]

    Will be converted into the C{TreeWidget} object::

       TreeWidget('IP', TreeWidget('DP', 'the', 'dog', roof=1),
                  TreeWidget('VP', TreeWidget('V', TreeWidget('V\'', 'saw',),
                             TreeWidget('DP', TreeWidget('D', 'the',),
                                        TreeWidget('NP', 'man')))))

    Note: this will not work for all valid qtree expressions
    (especially complex ones with quoted delimiters, etc).  
    
    @rtype: C{TreeWidget}
    """
    # Get rid of any newlines, etc.
    str = ' '.join(str.split())

    # Ignore several of LaTeX commands; they just clutter up the diagram.
    str = re.sub(r'!\\sqtw\{[^}]+\}', '', str)
    str = re.sub(r'\\node\{[^}]+\}', '', str)
    str = re.sub(r'\\text\w\w(?=\{)', '', str)
    str = re.sub(r'\\emph(?=\{)', '', str)
    
    # Recognize several LaTeX character commands.
    if REPLACE_LATEX_CHARS:
        str = replace_latex_chars(str)

    if str[0:5] != '\Tree':
        raise ValueError('Bad LaTeX-qtree string')
    str = str[5:]

    # Backslash any quote marks or backslashes in the string.
    str = re.sub(r'([\\"\'])', r'\\\1', str)
    
    ## Double-backslash is a newline
    #str = str.replace(r'\\\\', r'\n')

    (str, literals) = _extract_latex_literals(str)

    # Add null characters and commas around words.  We'll turn the
    # null characters into quotes later; but we want to be able to
    # distinguish our quotes from theirs.
    str = re.sub(r'([^\[\] ]+)', "\0\\1\0,", str)

    # Brackets delimit subtrees.
    str = re.sub("\\[\0\\.", "TreeWidget(\0", str)

    str = re.sub('\\[', "TreeWidget(\0\0,", str)
    str = re.sub("\\]\0\\.([^\0]+)\0", ').with_node(\0\\1\0)', str)
    str = re.sub('\\]', '),', str)

    # Handle qroofs.
    str = re.sub("\0\\\\\\\\qroof\\{(\\d+)\\}\\.([^\0]+)\0",
                 "TreeWidget(\0\\2\0, \0{\\1}\0, roof=1)", str)

    # Strip whitespace and get rid of the comma after the last ')' 
    str = str.strip()[:-1]

    # Substitute literals back in.  Don't include their delimiters.
    for i in range(len(literals)):
        str = str.replace(('{%d}' % i), literals[i])

    # Replace null characters with calls to parse-label
    str = re.sub('\0([^\0]*)\0', r"_parse_latex_qtree_label('\1')", str)

    # Use "eval" to convert the string (is this safe?)
    try:
        result = eval(str)
        return result
    except:
        raise ValueError('Bad LaTeX-qtree string')

def _parse_latex_qtree_label(str):
    # Hacks..
    str = re.sub(r'{({[^{}]*})}', r'\1', str)
    str = re.sub(r'([_^])(\\\w+)', r'\1{\2}', str)
    
    label = _parse_treebank_label(str)

    # A few final touches..
    for line in label.subwidgets:
        for text in line.subwidgets:
            text.str = text.str.replace('~', ' ')
            text.str = re.sub('[{}$]', '', text.str)

    return label

############################################################
##  Diagram Rendering
############################################################

class DiagramRenderer:
    """
    A processing class that converts diagrams into shape lists.
    """
    def __init__(self, font, xspace=1.5, yspace=1.5, linespace=1.0):
        """
        Create a new C{DiagramRenderer}.

        @param font: The font used to render text strings in the
            diagram.  This font is used to find font metric
            information.
        @type font: C{ImageFont.ImageFont}
        @type xspace: C{float}
        @param xspace: The number of characters space between nodes.
        @type yspace: C{float}
        @param yspace: The number of lines space between levels of the
            tree.
        @type linespace: C{float}
        @param linespace: Line spacing for multiline labels.
        """
        self._font = font
        self._xspace = xspace * font.getsize('M')[0]
        self._yspace = yspace * font.getsize('M')[1]
        self._linespace = linespace

    def _draw_text(self, text):
        """
        @return: A tuple (shapes, width, height)
        """
        (width, height) = self._font.getsize(text.str)
        if text.attribs.get('superscript', 0):
            return ([TextShape(0, 0, text.str)],
                    width, height*4/3)
        elif text.attribs.get('subscript', 0):
            return ([TextShape(0, height*1/3, text.str)],
                    width, height*4/3)
        else:
            return ([TextShape(0,0,text.str)], width, height)

    def _draw_sequence(self, sequence):
        """
        @return: A tuple (shapes, width, height)
        """
        # Use align to figure out how much we should shift the elements.
        align = sequence.attribs.get('align', 'center')
        if align == 'top': alignshift = 0.0
        elif align == 'center': alignshift = 0.5
        elif align == 'bottom': alignshift = 1.0
        else: raise ValueError('Bad alignment %s' % align)

        subwidget_draws = [self.draw(sw) for sw in sequence.subwidgets]
        height = max(0, 0, *[swd[2] for swd in subwidget_draws])
        width = 0
        all_elts = []
        for (elts, w, h) in subwidget_draws:
            for elt in elts: elt.shift(width, (height-h)*alignshift)
            all_elts += elts
            width += w

        return (all_elts, width, height)

    def _draw_stack(self, stack):
        """
        @return: A tuple (shapes, width, height)
        """
        # Use align to figure out how much we should shift the elements.
        align = stack.attribs.get('align', 'center')
        if align == 'left': alignshift = 0.0
        elif align == 'center': alignshift = 0.5
        elif align == 'right': alignshift = 1.0
        else: raise ValueError('Bad alignment %s' % align)
    
        subwidget_draws = [self.draw(sw) for sw in stack.subwidgets]
        width = max(0, 0, *[swd[1] for swd in subwidget_draws])
        height = 0
        all_elts = []
        for (elts, w, h) in subwidget_draws[:-1]:
            for elt in elts: elt.shift((width-w)*alignshift, height)
            all_elts += elts
            height += h * self._linespace
        for (elts, w, h) in subwidget_draws[-1:]:
            for elt in elts: elt.shift((width-w)*alignshift, height)
            all_elts += elts
            height += h
    
        return (all_elts, width, height)
    
    def _draw_tree(self, tree):
        """
        @return: A tuple (shapes, width, height, nodecenter).
            C{nodecenter} is the x coordinate of the tree's node's
            center (used to draw lines to subtrees).
        """
        # Draw the children.
        child_elts = []
        child_xs = []
        height = 0
        width = -self._xspace
        children = list(tree.children)
        children.reverse()
        for child in children:
            # If it's a tree, connect the line to its top node's center.
            if isinstance(child, TreeWidget):
                (elts, w, h, center) = self._draw_tree(child)
            else:
                (elts, w, h) = self.draw(child)
                center = w/2
    
            # Shift the other children right, and add the new child.
            for elt in child_elts: elt.shift(w + self._xspace, 0)
            child_elts += elts
    
            # Update width & height
            height = max(h, height)
            width += w + self._xspace
    
            # Update child_xs.
            child_xs = [x+w+self._xspace for x in child_xs]
            child_xs.append(center)
            
        # Draw the node label.
        (label_elts, w, h) = self.draw(tree.node)
        
        # Center the children with the node label.
        if w > width:
            for elt in child_elts: elt.shift((w-width)/2, 0)
            child_xs = [(w-width)/2+x for x in child_xs]
            width = w
    
        # Shift the children down to make room for the node label.
        for elt in child_elts: elt.shift(0, h+self._yspace)
        if len(child_elts) > 0: height += h+self._yspace
        else: height = h
    
        # Center the node label over its children (when possible)
        if len(child_xs) > 0:
            nodex = reduce(lambda a,b:a+b, child_xs, 0) / len(child_xs)
            if nodex + w/2 > width: nodex = width-w/2
            if nodex - w/2 < 0: nodex = w/2
        else:
            nodex = width/2
        for elt in label_elts: elt.shift(nodex-w/2, 0)
    
        # Draw lines from the node to the children.
        (y1, y2) = (h, h+self._yspace)
        if tree.attribs.get('roof', 0):
            line_elts = [LineShape(nodex, y1, width, y2),
                         LineShape(nodex, y1, 0, y2),
                         LineShape(0, y2, width, y2)]
        else:
            line_elts = [LineShape(nodex, y1, child_x, y2)
                         for child_x in child_xs]
    
        # These can be very useful for debugging:
        if DEBUG_DRAW:
            line_elts.append(LineShape(0,       1,        width-1, 1))
            line_elts.append(LineShape(width-1, 1,        width-1, 8))
            line_elts.append(LineShape(0,       1,        0,       8))
            line_elts.append(LineShape(width-1, height-1, width-1, height-6))
            line_elts.append(LineShape(width-1, height-1, width-5, height-1))
            line_elts.append(LineShape(0,       height-1, 0,       height-6))
            line_elts.append(LineShape(0,       height-1, 5,       height-1))
        
        return (line_elts+label_elts+child_elts, width, height, nodex)
    
    def draw(self, widget):
        """
        Convert a diagram widget to a shape list.

        @type widget: C{DiagramWidget}
        @param widget: The diagram widget that should be converted
            into a list of shapes.
        @return: A tuple C{(shapes, width, height)}, where C{shapes}
            is a list of the C{Shape}s that make up an image of the
            given diagram widget; C{width} is the width of the image;
            and C{height} is the height of the image.
        @rtype: (C{list} of C{Shape}s, C{int}, C{int})
        """
        if isinstance(widget, TreeWidget):
            return self._draw_tree(widget)[:3]
        elif isinstance(widget, SequenceWidget):
            return self._draw_sequence(widget)
        elif isinstance(widget, StackWidget):
            return self._draw_stack(widget)
        elif isinstance(widget, TextWidget):
            return self._draw_text(widget)            

############################################################
##  (temporary)
############################################################

def _draw_multiple_trees(trees, font, xspace, yspace, linespace):
    elts = []
    maxy = maxx = 0
    trees = list(trees)
    trees.reverse()
    for tree in trees:
        (new_elts, bbox, nodex) = _draw_tree(tree, font, xspace,
                                             yspace, linespace)
        for elt in elts: elt.shift(bbox[0]+10,0)
        elts += new_elts
        maxy = max(maxy, bbox[1])
        maxx += bbox[0] + 10
        
    return (elts, (maxx, maxy), maxx/2)

############################################################
##  Shape List to Raster Image Conversion
############################################################

def _string2color(s):
    if not re.match('^#[0-9a-fA-F]{6,6}$', s): raise ValueError()
    return (int(s[1:3],16), int(s[3:5],16), int(s[5:7],16))

def shapelist2image(shapes, width, height, font=DEFAULT_FONT, 
                    text_color=_string2color(DEFAULT_PS_TEXTCOLOR),
                    line_color=_string2color(DEFAULT_PS_LINECOLOR),
                    bg_color=_string2color(DEFAULT_PS_BGCOLOR)):
    """
    Build a PIL C{Image} containing the given list of C{Shape}s.

    @param font: The PIL C{ImageFont} used to draw the nodes
    @type font: C{string} or C{ImageFont.ImageFont}
    @param text_color: The RGB color used to draw the nodes
    @type text_color: C{tuple} of 3 C{int}s
    @param line_color: The RGB color used to draw lines
    @type line_color: C{tuple} of 3 C{int}s
    @param bg_color: The background color for the image.
    @rtype: C{Image.Image}
    """
    import Image, ImageDraw

    # If the font is specified as a string, then load it.
    if type(font) == type(''):
        import ImageFont
        font = ImageFont.load_path(font)
        
    # Create the (blank) image.
    im = Image.new('RGB', (width,height), color=bg_color)

    # Create an ImageDraw, to draw on the image.
    draw = ImageDraw.ImageDraw(im)

    # Draw the lines.
    lines = [s for s in shapes if isinstance(s, LineShape)]
    for line in lines:
        draw.line([(line.x1, line.y1), (line.x2, line.y2)], line_color)

    # Draw the text.
    texts = [s for s in shapes if isinstance(s, TextShape)]
    for text in texts:
        draw.text((text.x, text.y), text.text, text_color, font)
    return im

############################################################
##  Shape List to Postscript Image Conversion
############################################################

class PseudoFont:
    """
    Mimicks a PIL font, so we can do spacing calculations for
    postscript output.
    """
    def __init__(self, canvas, font):
        self._canvas = canvas
        self._font = font
    def getsize(self, str):
        # Bounding boxes are integer sized; repeat the string several
        # times to get a better estimate of its real size.
        txt = self._canvas.create_text(0,0,text=str*100,
                                       font=self._font,
                                       anchor='nw')
        bbox = self._canvas.bbox(txt)
        self._canvas.delete(txt)
        dx = (bbox[2]-bbox[0])/100.0
        dy = bbox[3]-bbox[1]
        # hack.. (Tkinter doesn't give us reliable #s for font sizes)
        if str == 't': dx *= 0.7
        return (dx,dy)

def shapelist2ps(shapes, width, height, canvas, output_file,
                 font=DEFAULT_PSFONT,
                 text_color=_string2color(DEFAULT_PS_TEXTCOLOR),
                 line_color=_string2color(DEFAULT_PS_LINECOLOR),
                 bg_color=_string2color(DEFAULT_PS_BGCOLOR)):
    """
    Use a Tkinter canvas to draw a list of C{Shape}s.
    """
    # Draw the background.
    if bg_color != (255, 255, 255):
        canvas.create_rectangle(0, 0, bbox[0], bbox[1],
                                fill=('#%02x%02x%02x' % bg_color),
                                outline=('#%02x%02x%02x' % bg_color))

    # Draw the lines.
    lines = [s for s in shapes if isinstance(s, LineShape)]
    for line in lines:
        canvas.create_line(line.x1, line.y1, line.x2, line.y2,
                           fill=('#%02x%02x%02x' % line_color))

    # Draw the text.
    texts = [s for s in shapes if isinstance(s, TextShape)]
    for text in texts:
        canvas.create_text(text.x, text.y+2, text=text.text,
                           fill=('#%02x%02x%02x' % text_color),
                           font=font, anchor='nw')

    # Print the tree to a file.
    canvas.postscript(file=output_file, width=width, height=height)

############################################################
##  Command-line Interface
############################################################

def _str2psfont(s):
    if re.match('^\d+$', s):
        return (DEFAULT_PSFONT[0], int(s))
    elif re.match('^[a-zA-Z]+$', s):
        return (s, DEFAULT_PSFONT[1])
    elif re.match('^\d+\s+[a-zA-Z]+$', s):
        (pt, name) = s.split()
        return (name, int(pt))
    elif re.match('^[a-zA-Z]+\s+\d+$', s):
        (name, pt) = s.split()
        return (name, int(pt))
    else:
        _usage()

def _usage():
    import os.path
    progname = os.path.basename(sys.argv[0])
    print
    print 'Usage: %s [-font n] [-xspace f] [-yspace f] [-type fmt]' % progname
    print ' '*(len(progname)+7),
    print '[-monochorome | -grayscale | -color] [-rv]'
    print ' '*(len(progname)+7),
    print '[-linecolor c] [-background c] [-textcolor c]'
    print ' '*(len(progname)+7),
    print '[-linespace f] <tree-file> <image-file> '
    print
    print '      -font n: A PIL font file, used to display the nodes.'
    print '               Default =', DEFAULT_FONT
    print '    -xspace f: The number of characters space between nodes.'
    print '               Default =', DEFAULT_XSPACE
    print '    -yspace f: The number of lines space between levels of the tree.'
    print '              Default =', DEFAULT_YSPACE
    print ' -linespace f: Line spacing for multiline labels.'
    print '               Default =', DEFAULT_LINESPACE
    print ' -linecolor c: Color for lines (Use #rrggbb format)'
    print '               Default =', DEFAULT_LINECOLOR
    print '-background c: Background color (Use #rrggbb format)'
    print '               Default =', DEFAULT_BGCOLOR
    print ' -textcolor c: Color for nodes (Use #rrggbb format)'
    print '               Default =', DEFAULT_TEXTCOLOR
    print '  -monochrome: Create a monochrome image',
    print '(default)' * (DEFAULT_IMAGECOLOR == 'monochrome')
    print '   -grayscale: Create a grayscale image',
    print '(default)' * (DEFAULT_IMAGECOLOR == 'grayscale')
    print '       -color: Create a color image',
    print '(default)' * (DEFAULT_IMAGECOLOR == 'color')
    print '    -type fmt: The format for the output image (e.g., "gif" or "png")'
    print '               Default = auto'
    print '          -rv: Invert the foreground and background colors.'
    print '  <tree-file>: A file containing paren or brace delimited treebank-'
    print '               style trees.'
    print ' <image-file>: The name of the image file to be created.  Should be a'
    print '               gif or a png file.'
    print
    sys.exit(-1)

def _error(str):
    import os.path
    print '%s error: %s' % (os.path.split(sys.argv[0])[1], str)
    sys.exit(-1)

def _log(str):
    import time, os.path
    print '%s: %s' % (os.path.split(sys.argv[0])[1], str)

def _main_do_postscript(tree, psfont, xspace, yspace, linespace,
                        textcolor, linecolor, bgcolor, verbosity,
                        image_filename):
    if verbosity >= 3: _log('  Creating Tkinter canvas')
    try: import Tkinter
    except Exception, e:
        _error("Postscript output requires Tkinter: %s" % e)
    top = Tkinter.Tk()
    canvas = Tkinter.Canvas(top)
    canvas.pack()

    if verbosity >= 3: _log('  Creating Tkinter psuedofont')
    pseudofont = PseudoFont(canvas, psfont)

    if verbosity >= 3: _log('  Rendering %r' % tree)
    elif verbosity >= 2: _log('  Rendering tree')
    try:
        renderer = DiagramRenderer(pseudofont, xspace, yspace, linespace)
        (shapes, width, height) = renderer.draw(tree)
    except Exception, e:
        _error("Could not render tree: %s" % e)

    if verbosity >= 3: _log('  Drawing %r as postscript' % (shapes,))
    elif verbosity >= 2: _log('  Drawing tree as postscript')
    try:
        im = shapelist2ps(shapes, width, height, canvas, image_filename, 
                          psfont, textcolor, linecolor, bgcolor)
    except Exception, e:
        _error("Could not draw image as postscript: %s" % e)

    # Is this necessary and/or useful?
    top.mainloop(1)
    
    # Clean up
    canvas.destroy()
    top.destroy()

    

def cli():
    """
    Run the command-line interface to tree2image.  Run the
    command-line interface with no arguments for usage.

    @rtype: C{None}
    """
    # ===============================================
    # Argument handling...
    # ===============================================
    linecolor = textcolor = bgcolor = reverse_video = psfont = None
    imagecolor = imagetype = tree_filename = image_filename = None
    fontfile = xspace = yspace = linespace = None
    verbosity = 0

    # Process all command-line arguments.
    args = sys.argv[1:]
    args.reverse() # reverse, so we can pop them off.
    while args:
        arg = args.pop()
        if arg[:1] == '-':
            if arg[:2] == '--': arg = arg[1:]
            arg = arg.lower()
            if arg in ('-f', '-font', '-fn'):
                if not args: _usage()
                fontfile = args.pop()
            elif arg in ('-psfont', '-psfn', '-psf'):
                if not args: _usage()
                psfont = _str2psfont(args.pop())
            elif arg in ('-x', '-xs', '-xspace', '-xspacing'):
                if not args: _usage()
                try: xspace = float(args.pop())
                except: _usage()
            elif arg in ('-textcolor', '-text'):
                if not args: _usage()
                try: textcolor = _string2color(args.pop())
                except: _usage()
            elif arg in ('-linecolor', '-line'):
                if not args: _usage()
                try: linecolor = _string2color(args.pop())
                except: _usage()
            elif arg in ('-bg', '-background', '-bgcolor'):
                if not args: _usage()
                try: bgcolor = _string2color(args.pop())
                except: _usage()
            elif arg in ('-y', '-ys', '-yspace', '-yspacing'):
                if not args: _usage()
                try: yspace = float(args.pop())
                except: _usage()
            elif arg in ('-l', '-ls', '-linespace', '-linespacing'):
                if not args: _usage()
                try: linespace = float(args.pop())
                except: _usage()
            elif arg in ('-help', '-h', '-?', '-usage'): _usage()
            elif arg in ('-v', '-verbose'): verbosity += 1
            elif arg in ('-vv', '-vvv'): verbosity += len(arg)-1
            elif arg in ('-reversevideo', '-rv', '-reverse', '-reverse_video'):
                reverse_video = not reverse_video
            elif arg in ('-mono', '-monochrome'):
                imagecolor = 'monochrome'
            elif arg in ('-gray', '-grayscale', '-grey', '-greyscale'):
                imagecolor = 'grayscale'
            elif arg in ('-color',):
                imagecolor = 'color'
            elif arg in ('-format', '-type', '-t', '-imagetype'):
                if not args: _usage()
                imagetype = args.pop().strip().lower()
            else: _usage()
        elif tree_filename is None:
            tree_filename = arg
        elif image_filename is None:
            image_filename = arg
        else: _usage()
    if image_filename is None: _usage()

    # Are we rendering postscript?
    postscript = (image_filename.endswith('.ps') or
                  image_filename.endswith('.eps'))

    # Assign defaults to everything else
    if postscript:
        if linecolor is None: linecolor = _string2color(DEFAULT_PS_LINECOLOR)
        if textcolor is None: textcolor = _string2color(DEFAULT_PS_TEXTCOLOR)
        if bgcolor is None: bgcolor = _string2color(DEFAULT_PS_BGCOLOR)
        if imagecolor is None: imagecolor = DEFAULT_PS_IMAGECOLOR
    else:
        if linecolor is None: linecolor = _string2color(DEFAULT_LINECOLOR)
        if textcolor is None: textcolor = _string2color(DEFAULT_TEXTCOLOR)
        if bgcolor is None: bgcolor = _string2color(DEFAULT_BGCOLOR)
        if imagecolor is None: imagecolor = DEFAULT_IMAGECOLOR
    
    if fontfile is None: fontfile = DEFAULT_FONT
    if xspace is None: xspace = DEFAULT_XSPACE
    if yspace is None: yspace = DEFAULT_YSPACE
    if linespace is None: linespace = DEFAULT_LINESPACE
    if imagetype is None: imagetype = 'auto'
    if verbosity is None: verbosity = DEFAULT_VERBOSITY
    if reverse_video is None: reverse_video = DEFAULT_REVERSE_VIDEO
    if psfont is None: psfont = DEFAULT_PSFONT

    # If the image type is monochrome, force monochrome.
    if imagetype in ('xbm',):
        imagecolor = 'monochrome'

    # If the image is monochrome, pick either white-on-black or black-on-white.
    if imagecolor == 'monochrome':
        if bgcolor == (255,255,255):
            textcolor = linecolor = (0,0,0)
        else:
            textcolor = linecolor = (255,255,255)
            bgcolor = (0,0,0)

    # Reverse video reverses colors.
    if reverse_video:
        textcolor = tuple([255-v for v in textcolor])
        linecolor = tuple([255-v for v in linecolor])
        bgcolor = tuple([255-v for v in bgcolor])

    if verbosity >= 1:
        _log('%s -> %s (%s %s)' % (tree_filename, image_filename,
                                  imagecolor, imagetype))
    if verbosity >= 3:
        _log('   Tree filename: %r' % tree_filename)
        _log('  Image filename: %r' % image_filename)
        _log('   Font filename: %r' % fontfile)
        _log('      Image type: %s %s' % (imagecolor, imagetype))
        _log('Background color: %s' % (bgcolor,))
        _log('      Line color: %s' % (linecolor,))
        _log('      Text color: %s' % (textcolor,))
        _log('       X Spacing: %f' % xspace)
        _log('       Y Spacing: %f' % yspace)

    # ===============================================
    # Main processing
    # ===============================================
    if verbosity >= 2: _log('  Reading %r' % tree_filename)
    try: tree_str = open(tree_filename).read()
    except Exception, e:
        _error("Could not read %r: %s" % (tree_filename, e))

    if verbosity >= 3: _log('  Parsing %r' % tree_str)
    elif verbosity >= 2: _log('  Parsing tree string')
    try: tree = parse_tree(tree_str)
    except Exception, e:
        _error("Error parsing %r: %s" % (tree_filename, e))

    # Special case for handling postscript. (what font name & size?)
    if postscript and SPECIAL_PS:
        _main_do_postscript(tree, psfont, xspace, yspace, linespace,
                            textcolor, linecolor, bgcolor, verbosity,
                            image_filename)
        return
        
    if verbosity >= 2: _log('  Loading font %r' % fontfile)
    try: import ImageFont
    except Exception, e:
        _error("Non-postscript output requires PIL: %s" % e)
    try: font = ImageFont.load_path(fontfile)
    except Exception, e:
        _error("Could not load font %r: %s" % (fontfile, e))

    if verbosity >= 3: _log('  Rendering %r' % tree)
    elif verbosity >= 2: _log('  Rendering tree')
    try:
        renderer = DiagramRenderer(font, xspace, yspace, linespace)
        (shapes, width, height) = renderer.draw(tree)
    except Exception, e:
        _error("Could not render tree: %s" % e)

    if verbosity >= 3: _log('  Drawing %r using PIL' % shapes)
    elif verbosity >= 2: _log('  Drawing tree using PIL')
    try:
        im = shapelist2image(shapes, width, height, font,
                             textcolor, linecolor, bgcolor)
    except ImportError, e:
        _error("Raster image output requires PIL: %s" % e)
    except Exception, e:
        _error("Could not create tree image: %s" % e)

    if verbosity >= 2: _log('  Converting image to %s' % imagecolor)
    if imagecolor == 'monochrome': im = im.convert('1')
    elif imagecolor == 'grayscale': im = im.convert('P')

    if verbosity >= 2: _log('  Saving to %r' % image_filename)
    try:
        if imagetype == 'auto': im.save(image_filename)
        else: im.save(image_filename, format=imagetype)
    except Exception, e:
        _error("Could not write to %s: %s" % (image_filename, e))

if __name__ == '__main__': cli()
